#!/usr/bin/perl
#
# Script to create maps for gpsdrive-ipaq
#
# Written by Andrew M. Bishop
#
# This file Copyright 2007, 2008 Andrew M. Bishop
# It may be distributed under the GNU Public License, version 2, or
# any higher version.  See section COPYING of the GNU Public license
# for conditions under which this file may be redistributed.

# Constants

$urlbase="http://tile.openstreetmap.org/mapnik";

$wget_options="";

$PI=3.14159265358979323846;

# Parse command line arguments

if($#ARGV < 2)
  {
   print "usage: gpsdrive-ipaq-osm.pl <scale> <lat> <lng>\n";
   print "usage: gpsdrive-ipaq-osm.pl <scale> <lat1> <lng1> <lat2> <lng2>\n";
   exit 1;
  }

$scale=$ARGV[0]+0;

die "Invalid scale value\n" if($scale<5 || $scale>16);

$lat1=$ARGV[1]+0;
$lng1=$ARGV[2]+0;

die "Invalid latitude\n"  if($lat1<-87.5 || $lat1>87.5);
die "Invalid longitude\n" if($lng1<-180 || $lng1>180);

if($#ARGV==4)
  {
   $lat2=$ARGV[3]+0;
   $lng2=$ARGV[4]+0;

   die "Invalid latitude\n"  if($lat2<-87.5 || $lat2>87.5);
   die "Invalid longitude\n" if($lng2<-180 || $lng2>180);
  }
else
  {
   $lat2=$lat1;
   $lng2=$lng1;
  }

# Find the x,y coordinates of the openstreetmap corner tiles

($x1,$y1)=ll_xy($scale,$lng1,$lat1);

($x2,$y2)=ll_xy($scale,$lng2,$lat2);

# Put them in the correct order

$tx1=min($x1,$x2);
$x2 =max($x1,$x2);
$x1=$tx1;

$ty1=min($y1,$y2);
$y2 =max($y1,$y2);
$y1=$ty1;

die "Cannot cross the line of -180 degrees longitude\n" if(($lng1<-90 && $lng2>90) || ($lng1>90 && $lng2<-90));

# Find the X,Y coordinates of the gpsdrive map (indexed by tile in north-west (top-left) corner)

$X1=int($x1/4);
$Y1=int($y1/3);

$X2=int($x2/4);
$Y2=int($y2/3);

print "Gpsdrive maps: $X1,$Y1 -> $X2,$Y2\n";

# Find the range of tiles to fetch

$x1=4*$X1;
$y1=3*$Y1;

$x2=4*$X2;
$y2=3*$Y2;

print "Need tiles   : $x1,$y1 -> $x2,$y2\n";

$numtiles=($x2-$x1+1)*($y2-$y1+1);

die "Too many tiles to fetch\n" if($numtiles>1000);

# Get the tiles

mkdir ".tilecache";

foreach $x ($x1..$x2)
  {
   foreach $y ($y1..$y2)
     {
      $url="$urlbase/$scale/$x/$y.png";
      $tile=".tilecache/tile_${scale}_${x}_${y}.png";

      if(! -f $tile || -z $tile)
        {
         print "Fetching $url -> $tile\n";

         system "wget -w 0 $wget_options $url -O $tile > /dev/null 2>&1";
        }
     }
  }

# Merge them

foreach $X ($X1..$X2)
  {
   foreach $Y ($Y1..$Y2)
     {
      $file="osm_${scale}_${X}_${Y}.gif";

      print "Creating gpsdrive map $X,$Y\n";

      if (! -f $file || -z $file)
        {
         system "convert -depth 8 -size 1280x1024 xc:white $file";
        }

      foreach $dx (0..4)
        {
         foreach $dy (0..3)
           {
            $x=4*$X+$dx;
            $y=3*$Y+$dy;

            $tile=".tilecache/tile_${scale}_${x}_${y}.png";

            if(-f $tile && -s $tile)
              {
               $posx=$dx*256;
               $posy=$dy*256;

               system "composite $tile -geometry +$posx+$posy -compose src-over $file $file";
              }
           }
        }
     }
  }

# Write the osm_koord.txt file

open(OSM_KOORD,">>osm_koord.txt");

foreach $X ($X1..$X2)
  {
   foreach $Y ($Y1..$Y2)
     {
      $file="osm_${scale}_${X}_${Y}.gif";

      ($lng,$lat)=xy_ll($scale,4*$X+2.5,3*$Y+2);

      printf OSM_KOORD "$file %.4f %.4f %d\n",$lat,$lng,$scale;
     }
  }

close(OSM_KOORD);

#
# Subroutines
#

sub tan { sin($_[0]) / cos($_[0]) }

sub min { if($_[0]<$_[1]) { $_[0]; } else { $_[1]; } }

sub max { if($_[0]>$_[1]) { $_[0]; } else { $_[1]; } }

sub ll_xy
{
 my($scale,$longitude,$latitude)=@_;

 $PI=3.14159265358979323846;

 my($x)=($longitude/180 + 1)/(2**(1-$scale));
 my($y)=(1 - log(&tan(($PI/4)*($latitude/90 + 1)))/$PI)/(2**(1-$scale));

 return (int($x),int($y));
}

sub xy_ll
{
 my($scale,$x,$y)=@_;

 $PI=3.14159265358979323846;

 my($longitude)=180*(($x * (2**(1-$scale)))-1);
 my($latitude)=90*((4/$PI)*atan2(exp($PI*(1-($y * (2**(1-$scale))))),1) - 1);

 return ($longitude,$latitude);
}
